#!/usr/bonsaitools/bin/perl -w

use lib '.';

use Date::Parse;
use strict;
use vars qw(%FORM %MFORM $vars $template);

require 'CGI.pl';

my $DEFAULT_DATE_FORMAT = '%Y-%m-%d %H:%i:%s';

#  %class is a mapping from single-character database field type to a
#  textual "class".  Field types in the same class keep custom field data
#  in the same database table.

my %class = (
    i => 'integer',
    d => 'date',
    t => 'date',
    s => 's_string',
    l => 'l_string',
    e => 'selection',
    m => 'selection',
);

#  %classinfo maps the classes defined above to a few bits of information.
#
#  'table':     The database table that stores bug data for the class
#  'column':    The column within the table where the bug data resides
#  'quote':     Boolean; true if field data should be quoted (via
#               SqlQuote()) when inserted into the table, false otherwise
#  'multiple':  Boolean; true if the field data is a list, false if the
#               data is a scalar
#
#  The 'column' entry may be a code reference; in that case, the referenced
#  subroutine should return a SQL expression which expresses the field data
#  in an appropriate format.  The subroutine will be passed a reference to a
#  hash which holds named parameters for the data-fetching operation.
#  (Dates are formatted on the server side using this scheme.)

my %classinfo = (
    integer  => {
        table    => 'cf_integer',
        column   => 'value',
        quote    => 0,
        multiple => 0,
    },
    date => {
        table    => 'cf_date',
        column   => \&date_column,
        quote    => 1,
        multiple => 0,
    },
    s_string => {
        table    => 'cf_shortstring',
        column   => 'value',
        quote    => 1,
        multiple => 0,
    },
    l_string => {
        table    => 'cf_longstring',
        column   => 'value',
        quote    => 1,
        multiple => 0,
    },
    selection => {
        table    => 'cf_selection',
        column   => 'label_id',
        quote    => 0,
        multiple => 1,
    },
);

#  The keys of %delta are the valid values for the various *-dir CGI
#  parameters.  The corresponding values are the appropriate offsets for
#  purposes of splicing.  For example, if @fields is an ordered array
#  of fields in some group, $base is the offset of the field relative to
#  which a new field $newfield is to be inserted, and $dir is 'before' or
#  'after', then the new field would be inserted into the proper place in
#  the field list by a statement like the following:
#
#  splice @fields, $base + $delta{$dir}, 0, $newfield;

my %delta = (
    before => 0,
    after  => 1,
);

#  The keys of %type are the valid values for the various *-type CGI
#  parameters.

my %type = map { $_ => undef } qw(string selection integer date);

#  %contextsub is a dispatch table.  The keys are the valid values for
#  the "context" CGI parameter.

my %contextsub = (
    'main-menu'    => \&handle_main_menu,
    'edit-field'   => \&edit_field,
    'rename-field' => \&rename_field,
    'create-field' => \&show_create_field,
    'remove-group' => \&remove_group,
    'remove-field' => \&remove_field,
);

#  %actionsub is a dispatch table.  The keys are the various values of
#  the "action" CGI parameter on the main menu page.

my %actionsub = (
    'Edit'   => \&show_edit,
    'Create' => \&show_create_field,
    'Insert' => \&insert_group,
    'Move'   => \&move_item,
    'Rename' => \&show_rename,
    'Remove' => \&show_remove,
);

############
#
#  Main program begins.
#

ConnectToDatabase();

quietly_check_login();

eval {
    main();
};

if ($@) {
    print "Content-Type: text/html\n\n";
    ThrowTemplateError($@);
}

exit;

#
#  Main program ends.
#
############

#
#  main()
#
#  Top-level dispatch subroutine.  If the "context" parameter is present,
#  dispatches to the appropriate subroutine based on it.  Otherwise
#  the main menu is shown if one of the "product", "group", or "global"
#  parameters is present, or the choose-scope page otherwise.
#

sub main {
    my $context = $FORM{'context'};

    if (!$context) {
        if (exists $FORM{'product'}
                || exists $FORM{'group'}
                || exists $FORM{'global'}) {
            show_main_menu();
        } else {
            show_choose_scope();
        }
        return;
    }

    defined(my $sub = $contextsub{$context})
        or die qq(Invalid context "$context"\n);

    $sub->();

}

#
#  process_template($name)
#
#  Prints a minimal CGI header, then processes and prints the template
#  in template/en/default/custom-field called "$name.html.tmpl".  Throws
#  a template error if the processing fails.
#

sub process_template {
    my $name = shift;

    print "Content-Type: text/html\n\n";

    $template->process("custom-field/$name.html.tmpl", $vars)
        or ThrowTemplateError($template->error());
}

#
#  read_form(@param)
#
#  Many routines need to know whether they're being invoked in the context
#  of a product or a group, and what the numeric value of the "product"
#  or "group" parameter is.  This routine returns a list containing first,
#  the string 'product' or 'group'; next, the value of the "product" or
#  "group" form parameter; next, the form parameters named by @param.
#

sub read_form {
    my @param = @_;
    my $scope;

    if (exists $FORM{'product'}) {
        $scope = 'product';
    } elsif (exists $FORM{'group'}) {
        $scope = 'group';
    } elsif (exists $FORM{'global'}) {
        die qq(Global scope not yet implemented\n);
    } else {
        die qq(No scope specified\n);
    }

    return ($scope, @FORM{$scope, @param});

}

#
#  handle_main_menu()
#
#  This routine simply dispatches to the appropriate action subroutine
#  depending on which submit button was pressed on the main menu.
#

sub handle_main_menu {
    my $action = $FORM{'action'};

    defined(my $sub = $actionsub{$action})
        or die qq(Invalid action "$action"\n);

    $sub->();

}

#
#  show_choose_scope()
#
#  This routine sets up the context for, and then displays, the
#  choose-scope template.
#

sub show_choose_scope {
    my (@products, @groups);

    process_query(qq{
        SELECT id, name
        FROM products
        ORDER BY name
    }, sub {
        push @products, { id => $_[0], name => $_[1] };
    });

    process_query(qq{
        SELECT group_id, group_name
        FROM cf_groups
        WHERE group_name IS NOT NULL
        ORDER BY group_name
    }, sub {
        push @groups, { id => $_[0], name => $_[1] };
    });

    $vars->{products} = \@products;
    $vars->{groups}   = \@groups;

    process_template('choose-scope');

}

#
#  show_main_menu()
#
#  This routine calls read_product() or read_group() based on whether
#  the "product" or "group" CGI parameters is set, then displays the
#  main-menu template.
#

sub show_main_menu {
    if (exists $FORM{'product'}) {
        read_product($FORM{'product'});
    } elsif (exists $FORM{'group'}) {
        read_group($FORM{'group'});
    } elsif (exists $FORM{'global'}) {
        die qq(Global scope not yet implemented\n);
    } else {
        die qq(Unknown scope\n);
    }
    process_template('main-menu');
}

#
#  insert_group()
#
#  This routine inserts a named group into a product.  The product is
#  identified by the 'product' CGI parameter, and the group by the
#  'insert-target' parameter; it will be inserted either before or after
#  the group identified by the 'insert-base' parameter, based on the
#  'insert-dir' parameter.  The main menu is redisplayed if all went well.
#

sub insert_group {
    my ($scope, $scope_id, $target, $dir, $base) =
        read_form('insert-target', 'insert-dir', 'insert-base');

    $scope eq 'product' or die qq(Groups may only be inserted into products\n);

    is_numeric($scope_id) or die qq(Invalid product ID "$scope_id"\n);
    is_numeric($target) or die qq(Invalid group ID "$target"\n);
    is_numeric($base) or die qq(Invalid group ID "$base"\n);

    defined(my $delta = $delta{$dir})
        or die qq(Invalid parameter "insert-dir"\n);

    # lock_tables(...);

    product_exists($scope_id) or die qq(Unknown product "$scope_id"\n);
    group_exists($target)     or die qq(Unknown group "$target"\n);

    my $order = reorder_members($scope => $scope_id);

    defined(my $base_index = $order->{index}->($base))
        or die qq(Group $base is not a member of product $scope_id\n);
    !defined($order->{index}->($target))
        or die qq(Product $scope_id already contains group $target\n);

    $order->{insert}->($base_index + $delta, $target);

    send_sql(qq{
        INSERT INTO cf_product_members (product_id, group_id, sortkey)
                                VALUES ($scope_id, $target, 0)
    });

    $order->{commit}->();

    my @bugs = simple_query(qq{
        SELECT b.bug_id
        FROM bugs b
        WHERE b.product_id = $scope_id
    });

    if (@bugs) {
        my @fields = read_group_data($target, date_format => undef);
        my %by_class = classify_fields(@fields);
        while (my ($class, $fields) = each %by_class) {
            my %hash = map { $$_{id} => $$_{default} } @$fields;
            insert_field_data($class, \%hash, @bugs);
        }
    }

    # unlock_tables();

    show_main_menu();

}

#
#  insert_field_data($class, $data, @bugs)
#
#  This routine inserts field data into a single database table.  $data
#  is a reference to a hash which associates field IDs (the keys) with field
#  data (the values).  All fields must of the same class, namely $class.  
#  All of the indicated fields in each of the bugs enumerated in @bugs will
#  have their fields set to the indicated values.  The string 'NULL' may
#  occur in @bugs, in which case the default values of the fields will be
#  set.
#

sub insert_field_data {
    my ($class, $data, @bugs) = @_;
    my $info = $classinfo{$class};
    my ($table, $column, $quote, $multiple) =
        @$info{qw(table column quote multiple)};
    $column = $column->({ date_format => undef }) if ref $column eq 'CODE';

    while (my ($field_id, $value) = each %$data) {
        my @value = $multiple ? @$value : $value;
        for my $val (@value) {
            if (defined $val) {
                $val = SqlQuote($val) if $quote;
            } else {
                $val = 'NULL';
            }
        }
        for my $bug_id (@bugs) {
            for my $val (@value) {
                send_sql(qq{
                    INSERT INTO $table (bug_id, field_id, $column)
                                VALUES ($bug_id, $field_id, $val)
                });
            }
        }
    }

}

#
#  move_item()
#
#  This routine moves the field identified by the CGI parameter
#  'move-target' to the 'move-dir' side of the field 'move-base'.  It then
#  redisplays the main menu.
#

sub move_item {
    my ($scope, $id, $target, $dir, $base) =
        read_form('move-target', 'move-dir', 'move-base');

    is_numeric($id) or die qq(Invalid $scope "$id"\n);

    defined(my $delta = $delta{$dir})
        or die qq(Invalid parameter "move-dir"\n);

    my $order = reorder_members($scope => $id);

    defined(my $source = $order->{index}->($target))
        or die "No such field $target\n";
    defined(my $dest = $order->{index}->($base))
        or die "No such field $base\n";

    ++$source if $source >= $dest;

    $order->{insert}->($dest + $delta, $target);
    $order->{remove}->($source);
    $order->{commit}->();

    show_main_menu();

}

#
#  show_edit()
#
#  This routine sets up the context for, and then displays, the edit-field
#  template.
#

sub show_edit {
    my ($scope, $scope_id, $target) = read_form('edit-target');

    is_numeric($scope_id) or die qq(Invalid $scope "$scope_id"\n);
    is_numeric($target)   or die qq(Invalid field ID "$target"\n);

    if ($scope eq 'product') {
        product_exists($scope_id)
            or die qq(Unknown product "$scope_id"\n);
    } elsif ($scope eq 'group') {
        group_exists($scope_id)
            or die qq(Unknown group "$scope_id"\n);
    }

    field_exists($target) or die qq(Unknown field "$target"\n);

    my $field = read_custom_fields($target);

    if ($field->{is_selection}) {
        die qq(Editing of selection fields not yet implemented\n);
    }

    $vars->{scope}    = $scope;
    $vars->{scope_id} = $scope_id;
    $vars->{field}    = $field;

    process_template('edit-field');

}

#
#  edit_field()
#
#  This routine is called when the administrator clicks the "Change Field"
#  button on the edit-field page.  It resets the default value of the field
#  identified by the "field" CGI parameter to the value of the
#  "new-default" parameter.  An exception is thrown if the new default
#  value isn't properly formatted.  Otherwise, the main menu is
#  redisplayed.
#

sub edit_field {
    my ($field, $default) = @FORM{'field', 'new-default'};

    is_numeric($field) or die qq(Invalid field ID "$field"\n);

    my $type = simple_query(qq{
        SELECT field_type
        FROM custom_fields
        WHERE field_id = $field
    });
    defined $type or die qq(Unknown field ID "$field"\n);

    defined(my $class = $class{$type})
        or die qq(Database corrupted: unknown type "$type"\n);

    my $table = $classinfo{$class}{table};

    if ($class eq 'integer') {
        $default = parse_integer($default);
    } elsif ($class eq 'date') {
        $default = parse_date($default, $type eq 't');
    } elsif ($class eq 's_string' || $class eq 'l_string') {
        $default = SqlQuote($default);
    } elsif ($class eq 'selection') {
        die qq(Editing of selection fields not yet implemented\n);
    }

    send_sql(qq{
        UPDATE $table
        SET value = $default
        WHERE bug_id IS NULL AND field_id = $field
    }); 

    show_main_menu();

}

#
#  show_rename()
#
#  This routine sets up the context for, and then displays, the
#  rename-field template.
#

sub show_rename {
    my ($scope, $scope_id, $field_id) = read_form('rename-target');
    my $field_name;

    is_numeric($field_id) or die qq(Invalid field ID "$field_id"\n);
    is_numeric($scope_id) or die qq(Invalid $scope ID "$scope_id"\n);

    process_query(qq{
        SELECT display_name
        FROM custom_fields
        WHERE field_id = $field_id
    }, sub {
        $field_name = $_[0];
    }) or die qq(No such field "$field_id"\n);

    $vars->{field_name} = $field_name;
    $vars->{field_id}   = $field_id;
    $vars->{scope}      = $scope;
    $vars->{scope_id}   = $scope_id;

    process_template('rename-field');

}

#
#  rename_field()
#
#  This routine sets the name of the field identified by the "target-field"
#  CGI parameter to the value of the "new-name" parameter, then redisplays
#  the main menu.
#

sub rename_field {
    my ($field, $name) = @FORM{'target-field', 'new-name'};

    is_numeric($field)   or die qq(Invalid field ID "$field"\n);
    field_exists($field) or die qq(Unknown field ID "$field"\n);

    # lock_tables(...);

    valid_display_name($name)
        or die qq(Invalid field display name "$name"\n);

    $name = SqlQuote($name);

    send_sql(qq{
        UPDATE custom_fields
        SET display_name = $name
        WHERE field_id = $field
    });

    # unlock_tables();

    show_main_menu();

}

#
#  is_numeric($str)
#
#  Returns true if $str is of nonzero length and consists solely of numeric
#  characters, and false otherwise.
#

sub is_numeric {
    my $str = shift;
    $str =~ /^\d+$/;
}

#
#  valid_field_name($name)
#
#  Returns true if $name is a valid field name, and false otherwise.  A
#  valid field name starts with a non-numeric word character, and is
#  followed by zero or more word characters.
#

sub valid_field_name {
    my $name = shift;
    $name =~ /^[^\W\d]\w*/;
}

#
#  valid_display_name($name)
#
#  Returns true if $name is a valid display name, and false otherwise.  A
#  valid display name does not begin or end with whitespace, and does not
#  contain any vertical whitespace.
#

sub valid_display_name {
    my $name = shift;
    $name !~ /^\s/ && $name !~ /\s$/ && $name !~ /[\n\r]/;
}

#
#  validate_initial_create_params($param)
#
#  This routine validates the 'create-type', 'create-dir', 'create-base',
#  and 'progress' CGI parameters, which are present throughout the field
#  creation process.  The parameters are stored in the hash referenced by
#  the $param argument, under the keys 'type', 'dir', 'base', and
#  'progress', respectively.  In addition, $param->{scope} is set to either
#  'product' or 'group', and $param->{scope_id} is set to the value of the
#  'product' or 'group' CGI parameter.
#

sub validate_initial_create_params {
    my $param = shift;
    my ($scope, $scope_id, $type, $dir, $base, $progress) =
        read_form('create-type', 'create-dir', 'create-base', 'progress');

    exists $type{$type}
        or die qq(Invalid parameter "create-type"\n);
    !defined $dir || exists $delta{$dir}
        or die qq(Invalid parameter "create-dir"\n);
    !defined $base || is_numeric($base)
        or die qq(Invalid parameter "create-base"\n);

    if ($progress eq "") {
        $progress = 0;
    } elsif ($progress !~ /^\d$/) {
        die qq(Invalid parameter "progress"\n);
    }

    if ($scope eq 'product') {
        is_numeric    ($scope_id) or die qq(Invalid product ID "$scope_id"\n);
        product_exists($scope_id) or die qq(No such product "$scope_id"\n);
        !defined $base || product_has_group($scope_id, $base)
            or die qq(No such group $base in product $scope_id\n);
    } elsif ($scope eq 'group') {
        is_numeric($scope_id) or die qq(Invalid group ID "$scope_id"\n);
        group_exists($scope_id) or die qq(No such group "$scope_id"\n);
        !defined $base || group_has_field($scope_id, $base)
            or die qq(No such field $base in group $scope_id\n);
    }

    @$param{qw(scope scope_id progress type dir base)} =
        ($scope, $scope_id, $progress, $type, $dir, $base);

}

#
#  validate_secondary_create_params($param)
#
#  This routine validates the CGI parameters which appear on the first
#  create-field page.  These always include 'field-name', 'display-name',
#  and 'required', and others that depend on the type of the field being
#  created:  'default' (for integer and date fields), 'with-time' (for
#  date fields), 'size' (for string fields), and/or 'multi', 'unset', and
#  'domain' (for selection fields).  If valid, these parameters are stored
#  in the hash referenced by the $param argument using shorter, more
#  convenient names.
#

sub validate_secondary_create_params {
    my $param = shift;
    my ($name, $dname, $required) =
        @FORM{'field-name', 'display-name', 'required'};

    valid_field_name($name)    or die qq(Invalid field name "$name"\n);
    valid_display_name($dname) or die qq(Invalid display name "$dname"\n);
    $required = $required ? 1 : 0;

    @$param{qw(name dname required)} = ($name, $dname, $required);

    my $quoted = SqlQuote($name);
    0 == process_query(qq{
        SELECT 1
        FROM custom_fields
        WHERE field_name = $quoted
    }) or die qq(Custom field "$name" already exists\n);

    if ($param->{type} eq 'integer') {
        my $default = $FORM{'default'};
        if ($default !~ /\S/) {
            $default = 'NULL';
        } else {
            $default =~ s/^\s+//;
            $default =~ s/\s+$//;
            $default =~ /^[-+]?\d+$/
                or die qq(Invalid integer value "$default"\n);
        }
        $param->{default} = $default;
    } elsif ($param->{type} eq 'date') {
        my ($default, $withtime) = @FORM{'default', 'with-time'};
        $withtime = $withtime ? 1 : 0;
        $default = parse_date($default, $withtime);
        @$param{qw(with_time default)} = ($withtime, $default);
    } elsif ($param->{type} eq 'string') {
        my $size = $FORM{'size'};
        $size eq 'short' || $size eq 'long'
            or die qq(Invalid parameter "size"\n);
        $param->{size} = $size;
    } elsif ($param->{type} eq 'selection') {
        my ($multi, $unset, $domain) = @FORM{'multi', 'unset', 'domain'};
        $multi eq 'yes' || $multi eq 'no'
            or die qq(Invalid parameter "multi"\n);
        $unset =~ /^[^\r\n]+$/
            or die qq(Invalid parameter "unset"\n);
        if ($param->{progress} == 1) {
            my %seen = ($unset => 1);
            $domain =~ s/^\s+//gm;
            $domain =~ s/\s+$//gm;
            $domain =~ s/\n+/\n/g;
            $domain !~ /\r/ or die qq(Invalid parameter "domain"\n);
            $domain =~ /\S/ or die qq(Invalid parameter "domain"\n);
            $domain = [ grep !$seen{$_}++, split /\n/, $domain ];
        } else {
            $domain = [ deserialize_array($domain) ];
        }
        @$param{qw(multi unset domain)} = ($multi, $unset, $domain);
    }

}

#
#  validate_tertiary_create_params($param)
#
#  This routine validates the CGI parameters which appear on the second
#  create-field page (which is only reached when a string or selection
#  field is being created).  The only parameter entered on this page in
#  both cases is 'default', which is stored in $param->{default}.
#

sub validate_tertiary_create_params {
    my $param = shift;
    my $default;

    if ($param->{type} eq 'string') {
        $default = $FORM{'default'};
        if ($param->{size} eq 'short') {
            $default !~ /[\r\n]/
                or die qq(Invalid short string value "$default"\n);
        }
    } elsif ($param->{type} eq 'selection') {
        my %domain;
        @domain{ @{ $param->{domain} } } = ();
        $default = $MFORM{'default'};
        if ($param->{multi} eq 'no') {
            @$default == 1 or die qq(Single-selection fields can have ) .
                                  qq(only a single default value\n);
            my $label = $default->[0];
            if ($label eq $param->{unset}) {
                $default = undef;
            } else {
                exists $domain{$label}
                    or die qq(Invalid default selection value "$label"\n);
            }
        } else {
            for my $value (@$default) {
                exists $domain{$value}
                    or die qq(Invalid default selection value "$value"\n);
            }
        }
    } else {
        die qq(Invalid parameter "progress"\n);
    }

    $param->{default} = $default;

}

#
#  show_create_field()
#
#  This routine sets up the context for, and then displays, the
#  create-field template.  If all of the necessary information to create a
#  field has been supplied, then the field is created and the main menu is
#  redisplayed instead.
#

sub show_create_field {
    my (%param, %hidden);

    validate_initial_create_params(\%param);
    validate_secondary_create_params(\%param) if $param{progress} >= 1;
    validate_tertiary_create_params(\%param)  if $param{progress} >= 2;

    if ($param{progress} == 0) {
        @hidden{$param{scope}, 'create-type'} = @param{qw(scope_id type)};
        @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
            if defined $param{dir} && defined $param{base};
        @$vars{qw(hidden create_type progress)} = (\%hidden, $param{type}, 1);
        process_template('create-field');
        return;
    }

    if ($param{progress} == 1) {
        if ($param{type} eq 'integer' || $param{type} eq 'date') {
            create_field(\%param);
            show_main_menu();
            return;
        } elsif ($param{type} eq 'string') {
            @hidden{$param{scope}, qw(
                create-type field-name display-name required size
            )} = @param{qw(
                scope_id type name dname required size
            )};
            @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
                if defined $param{dir} && defined $param{base};
            @$vars{qw(hidden create_type size progress)} =
                (\%hidden, $param{type}, $param{size}, 2);
            process_template('create-field');
            return;
        } elsif ($param{type} eq 'selection') {
            @hidden{$param{scope}, qw(
                create-type field-name display-name required multi
                unset domain
            )} = (@param{qw(
                scope_id type name dname required multi unset
            )}, serialize_array($param{domain}));
            @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
                if defined $param{dir} && defined $param{base};
            @$vars{qw(hidden create_type multi domain progress)} =
                (\%hidden, @param{qw(type multi domain)}, 2);
            $vars->{unset} = $param{unset} if $param{multi} eq 'no';
            process_template('create-field');
            return;
        }
    }

    if ($param{progress} == 2) {
        create_field(\%param);
        show_main_menu();
        return;
    }
    
    die qq(Invalid parameter "progress"\n);

}

#
#  product_exists($product)
#
#  Returns true if the product identified by the numeric ID $product
#  exists, and false otherwise.
#

sub product_exists {
    my $product = shift;
    return process_query(qq{
        SELECT 1
        FROM products
        WHERE id = $product
    });
}

#
#  group_exists($group)
#
#  Returns true if the group identified by the numeric ID $group exists,
#  and false otherwise.
#

sub group_exists {
    my $group = shift;
    return process_query(qq{
        SELECT 1
        FROM cf_groups
        WHERE group_id = $group
    });
}

#
#  field_exists($field)
#
#  Returns true if the field identified by the numeric ID $field exists,
#  and false otherwise.
#

sub field_exists {
    my $field = shift;
    return process_query(qq{
        SELECT 1
        FROM custom_fields
        WHERE field_id = $field
    });
}

#
#  product_has_group($product, $group)
#
#  Returns true if the product identified by the numeric ID $product
#  contains a group with numeric ID $group, and false otherwise.
#

sub product_has_group {
    my ($product, $group) = @_;
    return process_query(qq{
        SELECT 1
        FROM cf_product_members
        WHERE product_id = $product AND group_id = $group
    });
}

#
#  group_has_field($group, $field)
#
#  Returns true if the group identified by the numeric ID $group contains
#  the field with numeric ID $field, and false otherwise.
#

sub group_has_field {
    my ($group, $field) = @_;
    return process_query(qq{
        SELECT 1
        FROM cf_group_members
        WHERE group_id = $group AND field_id = $field
    });
}

#
#  serialize_array($aref)
#
#  Returns a string containing a serialized representation of the array
#  referenced by $aref.  "|" characters within each element of the array
#  are doubled, and each element is then joined together with "|"
#  characters.  This is useful for storing array quantities (for example,
#  the "domain" parameter of a selection field) in a hidden CGI element.
#
#  Example: serialize_array(['abc', 'd|e', 'fgh']) returns 'abc|d||e|fgh'.
#

sub serialize_array {
    my $aref = shift;
    my @array = @$aref;
    s/\|/||/g for @array;
    return join '|', @array;
}

#
#  deserialize_array($ser)
#
#  Undoes the serialization performed by serialize_array().
#

sub deserialize_array {
    my $ser = shift;
    my $elem = "";
    my @array;
    while ($ser =~ /(.*?)\|/gcs) {
        $elem .= $1;
        if ($ser =~ /\G\|/gc) {
            $elem .= '|';
        } else {
            push @array, $elem;
            $elem = "";
        }
    }
    push @array, $elem . substr($ser, pos($ser));
    return @array;
}

#
#  create_field($param)
#
#  This routine creates a custom field.  All of the necessary information
#  must be present in the hash referenced by the $param argument under the
#  following keys:
#
#    'name':      The name of the field.
#    'dname':     The display name of the field.
#    'base':      The ID of the field relative to which the new field will be
#                 placed.
#    'dir':       Which side of the field 'base' the new field should be
#                 placed on, either 'before' or 'after'.
#    'required':  True if the new field is required on field creation,
#                 false otherwise.
#    'type':      The major type of the new field, either 'integer',
#                 'date', 'string', or 'selection'.
#    'default':   The default value of the field.
#    'with_time': (date fields only)  True if the field should include a
#                 time-of-day, false otherwise.
#    'size':      (string fields only)  'long' if the field should be a
#                 long string field, 'short' otherwise.
#    'multi':     (selection fields only)  'yes' if the field should be a
#                 multiselection field, 'no' otherwise.
#    'unset':     (selection fields only)  The label that should be used to
#                 identify a field with no label selected.
#    'domain':    (selection fields only)  A reference to an array
#                 containing all of the valid labels for the new field.
#
#  All of these values are presumed to have already been validated.
#

sub create_field {
    my $param = shift;
    my ($bugtable, $label_id, $dbtype, $field_id, $newindex, @bugs);
    my $selection_id = 'NULL';

    my ($scope, $scope_id, $base, $dir, $required, $name, $dname,
        $type, $default) = @$param{qw(
        scope scope_id base dir required name dname type default
    )};

    if ($type eq 'integer') {
        $dbtype = 'i';
        $bugtable = 'cf_integer';
    } elsif ($type eq 'date') {
        $dbtype = $param->{with_time} ? 't' : 'd';
        $bugtable = 'cf_date';
    } elsif ($type eq 'string') {
        $dbtype = $param->{size} eq 'short' ? 's' : 'l';
        $bugtable = $dbtype eq 's' ? 'cf_shortstring' : 'cf_longstring';
    } elsif ($type eq 'selection') {
        $dbtype = $param->{multi} eq 'no' ? 'e' : 'm';
        $bugtable = 'cf_selection';
    }

    # lock_tables( ... );

    my $order = reorder_members($scope => $scope_id);

    if (defined $base && defined $dir) {
        defined($order->{index}->($base)) or die qq(No such field "$base"\n);
        $newindex = $order->{index}->($base) + $delta{$dir};
    } else {
        $newindex = 0;
    }

    if ($type eq 'selection') {
        ($selection_id, $label_id) =
            create_selection_metadata(@$param{'unset', 'domain'});
        @$default = @$label_id{ @$default };
    }

    my $class = $class{$dbtype};

    $name   = SqlQuote($name);
    $dname  = SqlQuote($dname);
    $dbtype = SqlQuote($dbtype);

    send_sql(qq{
        INSERT INTO custom_fields (
            field_id, field_name, display_name, field_type,
            required, selection_id
        ) VALUES (
            NULL, $name, $dname, $dbtype, $required, $selection_id
        )
    });
    $field_id = last_insert_id();

    $order->{insert}->($newindex, $field_id);

    if ($scope eq 'product') {
        my $group_id = create_singleton_group($field_id);
        send_sql(qq{
            INSERT INTO cf_product_members (product_id, group_id, sortkey)
                                    VALUES ($scope_id, $group_id, 0)
        });
        @bugs = simple_query(qq{
            SELECT bug_id
            FROM bugs
            WHERE product_id = $scope_id
        });
    } elsif ($scope eq 'group') {
        send_sql(qq{
            INSERT INTO cf_group_members (group_id, field_id, sortkey)
                                  VALUES ($scope_id, $field_id, 0)
        });
        @bugs = simple_query(qq{
            SELECT b.bug_id
            FROM bugs b
              INNER JOIN cf_product_members pm USING (product_id)
            WHERE pm.group_id = $scope_id
        });
    }

    $order->{commit}->();

    insert_field_data($class, { $field_id => $default }, 'NULL', @bugs);

    # unlock_tables();

}

#
#  create_singleton_group($field_id)
#
#  Creates an unnamed group containing the single field identified by the
#  numeric ID $field_id.  Returns the new group's numeric ID.
#

sub create_singleton_group {
    my $field_id = shift;

    send_sql(qq{
        INSERT INTO cf_groups (group_id, group_name)
                       VALUES (NULL, NULL)
    });
    my $group_id = last_insert_id();

    send_sql(qq{
        INSERT INTO cf_group_members (group_id, field_id, sortkey)
                              VALUES ($group_id, $field_id, 1)
    });

    return $group_id;

}

#
#  create_selection_metadata($unset, $domain)
#
#  Populates the cf_selections and cf_selection_labels tables with
#  information for a new selection field.  Returns a two-element list.  The
#  first element is the numeric ID of the new selection set; the second is
#  a reference to hash which maps the new labels to the numeric ID of each.
#

sub create_selection_metadata {
    my ($unset, $domain) = @_;
    my ($n, $selection_id, %label_id);

    $unset = SqlQuote($unset);

    send_sql(qq{
        INSERT INTO cf_selections (selection_id, unset_label)
                           VALUES (NULL, $unset)
    });
    $selection_id = last_insert_id();

    $n = 1;
    for my $label (@$domain) {
        my $quoted = SqlQuote($label);
        send_sql(qq{
            INSERT INTO cf_selection_labels (
                label_id, selection_id, sortkey, label
            ) VALUES (
                NULL, $selection_id, $n, $quoted
            )
        });
        $label_id{$label} = last_insert_id();
        ++$n;
    }

    return ($selection_id, \%label_id);

}

#
#  show_remove()
#
#  Dispatches to either show_remove_from_product or show_remove_from_group,
#  depending on whether the 'product' or 'group' CGI parameter is present.
#

sub show_remove {
    my ($scope, $scope_id, $target) = read_form('remove-target');

    if ($scope eq 'product') {
        show_remove_from_product($scope_id, $target);
    } elsif ($scope eq 'group') {
        show_remove_from_group($scope_id, $target);
    }

}

#
#  show_remove_from_product($product, $target)
#
#  This routine sets up the context for, and then displays, the
#  remove-from-product template.  $target is the numeric ID of a group
#  to be removed from the product identified by the numeric ID $product.
#

sub show_remove_from_product {
    my ($product, $target) = @_;
    my ($product_name, $name, $object);

    is_numeric($product) or die qq(Invalid product "$product"\n);
    is_numeric($target)  or die qq(Invalid group "$target"\n);

    process_query(qq{
        SELECT name
        FROM products
        WHERE id = $product
    }, sub {
        $product_name = $_[0];
    }) or die qq(No such product "$product"\n);

    process_query(qq{
        SELECT group_name
        FROM cf_groups
        WHERE group_id = $target
    }, sub {
        $name = $_[0];
    }) or die qq(Unknown group "$target"\n);

    if (defined $name) {
        $object = 'group';
    } else {
        $object = 'field';
        process_query(qq{
            SELECT cf.display_name
            FROM custom_fields cf
              INNER JOIN cf_group_members gm USING (field_id)
            WHERE gm.group_id = $target
        }, sub {
            $name = $_[0];
        }) or die qq(Unknown field "$target"\n); #  DB corrupt!
    }

    $vars->{bugs_affected} = simple_query(qq{
        SELECT COUNT(*)
        FROM bugs
        WHERE product_id = $product
    });

    $vars->{product}      = $product;
    $vars->{product_name} = $product_name;
    $vars->{target_name}  = $name;
    $vars->{target_id}    = $target;
    $vars->{object}       = $object;

    process_template('remove-from-product');

}

#
#  show_remove_from_group($group_id, $field_id)
#
#  This routine sets of the context for, and then displays, the
#  remove-from-group template.  $field_id is the numeric ID of a field to
#  be removed from the group identified by the numeric ID $group_id.
#

sub show_remove_from_group {
    my ($group_id, $field_id) = @_;
    my ($group_name, $field_name, $bug_count, $product_count);

    is_numeric($group_id) or die qq(Invalid group ID "$group_id"\n);
    is_numeric($field_id) or die qq(Invalid field ID "$field_id"\n);

    process_query(qq{
        SELECT group_name
        FROM cf_groups
        WHERE group_id = $group_id
    }, sub {
        $group_name = $_[0];
    }) or die qq(No such group ID "$group_id"\n);

    defined $group_name or die qq(Cannot remove group "$group_id"\n);
    #  (Because it's private to some product.)

    process_query(qq{
        SELECT display_name
        FROM custom_fields
        WHERE field_id = $field_id
    }, sub {
        $field_name = $_[0];
    }) or die qq(No such field ID "$field_id"\n);

    $bug_count = simple_query(qq{
        SELECT COUNT(*)
        FROM bugs
          INNER JOIN cf_product_members pm USING (product_id)
        WHERE pm.group_id = $group_id
    });

    $product_count = simple_query(qq{
        SELECT COUNT(*)
        FROM cf_product_members
        WHERE group_id = $group_id
    });

    $vars->{group}         = $group_id;
    $vars->{group_name}    = $group_name;
    $vars->{target_id}     = $field_id;
    $vars->{target_name}   = $field_name;
    $vars->{bugs_affected} = $bug_count;
    $vars->{product_count} = $product_count;

    process_template('remove-from-group');

}

sub lock_tables {
    # ...
    return;
}

sub unlock_tables {
    send_sql("UNLOCK TABLES");
}

#
#  remove_bug_data($bugs, $meta)
#
#  This routines deletes all custom field data associated with each of the
#  bugs whose numeric IDs are contained in the array referenced by the
#  $bugs argument.  $meta is a field metadata object, as returned by
#  read_field_metadata().
#

sub remove_bug_data {
    my ($bugs, $meta) = @_;
    return if @$bugs == 0;
    my $bug_set = join(',', @$bugs);

    my @trans = simple_query(qq{
        SELECT transaction_id
        FROM cf_transactions
        WHERE bug_id IN ($bug_set)
    });
    my $trans_set = join(',', @trans);

    while (my ($table, $fields) = $meta->{each_table}->()) {
        my $field_set = join(',', @$fields);
        send_sql(qq{
            DELETE FROM $table
            WHERE bug_id IN ($bug_set)
              AND field_id IN ($field_set)
        });
        if (@trans) {
            send_sql(qq{
                DELETE FROM ${table}_activity
                WHERE transaction_id IN ($trans_set)
            });
        }
    }

    if (@trans) {
        send_sql(qq{
            DELETE FROM cf_transactions
            WHERE transaction_id IN ($trans_set)
        });
    }

}

#
#  remove_field_metadata($group, $meta)
#
#  This routine removes all metadata describing each of the fields
#  described by $meta (a field metadata object as returned by
#  read_field_metadata()) from the group identified by the numeric ID
#  $group.  The remaining fields in that group are re-ordered to close the
#  gap.
#

sub remove_field_metadata {
    my ($group, $meta) = @_;
    my @fields;

    while (my ($table, $fields) = $meta->{each_table}->()) {
        push @fields, @$fields;
        my $field_set = join(',', @$fields);
        send_sql(qq{
            DELETE FROM $table
            WHERE bug_id IS NULL
              AND field_id IN ($field_set)
        });
        if ($table eq 'cf_selection') {
            my $selection_set = join(',', $meta->{selection_id}->(@$fields));
            send_sql(qq{
                DELETE FROM cf_selection_labels
                WHERE selection_id IN ($selection_set)
            });
            send_sql(qq{
                DELETE FROM cf_selections
                WHERE selection_id IN ($selection_set)
            });
        }
    }

    my $order = reorder_members(group => $group);

    #  Sort into sortkey-order:
    @fields = sort { $order->{index}->($a) <=> $order->{index}->($b) } @fields;

    for (my $i = 0; $i < @fields; ++$i) {
        $order->{remove}->($order->{index}->($fields[$i]) - $i);
    }

    $order->{commit}->();

    my $field_set = join(',', @fields);

    send_sql(qq{
        DELETE FROM cf_group_members
        WHERE group_id = $group
          AND field_id IN ($field_set)
    });

    send_sql(qq{
        DELETE FROM custom_fields
        WHERE field_id IN ($field_set)
    });

}

#
#  remove_field()
#
#  This routine removes all bug data and field metadata for the field
#  identified by the numeric ID $FORM{'target'} from the group identified
#  by the numeric ID $FORM{'group'}.  It then redisplays the main menu.
#

sub remove_field {
    my ($group, $field) = @FORM{'group', 'target'};

    is_numeric($group) or die qq(Invalid group ID "$group"\n);
    is_numeric($field) or die qq(Invalid field ID "$field"\n);

    #  lock_tables(...);

    my @bugs = simple_query(qq{
        SELECT b.bug_id
        FROM bugs b
          INNER JOIN cf_product_members pm USING (product_id)
          INNER JOIN cf_group_members gm USING (group_id)
        WHERE gm.field_id = $field
    });

    my $meta = read_field_metadata($field);

    remove_bug_data(\@bugs, $meta);

    remove_field_metadata($group, $meta);

    # unlock_tables();

    show_main_menu();

}

#
#  remove_group()
#
#  This routine removes the group identified by the numeric ID
#  $FORM{'target'} from the product identified by the numeric ID
#  $FORM{'product'}.  The remaining groups in that product are re-ordered
#  to close the gap.  Bug data for all of the fields in the group is
#  removed.  If the group is anonymous--that is, if it contains only a
#  single field which belongs to a single product--the contained field is
#  removed as well.  Finally, the main menu is redisplayed.
#

sub remove_group {
    my ($product, $group) = @FORM{'product', 'target'};
    my $group_name;

    is_numeric($product) or die qq(Invalid product ID "$product"\n);
    is_numeric($group)   or die qq(Invalid group ID "$group"\n);

    product_exists($product) or die qq(Unknown product ID "$product"\n);
    group_exists($group)     or die qq(Unknonw group ID "$group"\n);

    # lock_tables(...);

    is_numeric($group) or die qq(Invalid group ID "$group"\n);

    process_query(qq{
        SELECT group_name
        FROM cf_groups
        WHERE group_id = $group
    }, sub {
        $group_name = $_[0];
    }) or die qq(Unknown group "$group"\n);

    my $order = reorder_members(product => $product);
    $order->{remove}->($order->{index}->($group));
    $order->{commit}->();

    my @bugs = simple_query(qq{
        SELECT b.bug_id
        FROM bugs b
          INNER JOIN cf_product_members pm USING (product_id)
        WHERE pm.group_id = $group
    });

    my @fields = simple_query(qq{
        SELECT field_id
        FROM cf_group_members
        WHERE group_id = $group
    });

    my $meta = read_field_metadata(@fields);

    remove_bug_data(\@bugs, $meta);

    send_sql(qq{
        DELETE FROM cf_product_members
        WHERE product_id = $product
          AND group_id = $group
    });

    if (!defined $group_name) {
        remove_field_metadata($group, $meta);
        send_sql(qq{
            DELETE FROM cf_groups
            WHERE group_id = $group
        });
    }

    # unlock_tables();

    show_main_menu();

}

#
#  send_sql($sql)
#
#  This routine massages the SQL statement contained in $sql to make it
#  more suitable for logging purposes, then passes it on to SendSQL().
#  Leading and trailing whitespace is removed, and all newline-indentation
#  sequences are collapsed to a single space.
#

sub send_sql {
    my $sql = shift;
    $sql = trim($sql);
    $sql =~ s/\n\s*/ /g;
    SendSQL($sql);
}

#
#  process_query($sql [, $sub])
#
#  This routine iterates over the result set obtained by passing $sql to
#  send_sql(), passing each row in the set to the subroutine $sub.  The
#  number of rows in the result set is returned.
#

sub process_query {
    my ($sql, $sub) = @_;
    $sub ||= sub { };  # default to no-op
    my $rows = 0;

    send_sql($sql);

    while (MoreSQLData()) {
        $sub->(FetchSQLData());
        ++$rows;
    }

    return $rows;

}

#
#  simple_query($sql)
#
#  This routine is useful for queries which return only a single column.
#  It returns a list containing the first column of each row in the result
#  set of the SQL statement $sql.
#

sub simple_query {
    my $sql = shift;
    my @result;

    send_sql($sql);

    while (MoreSQLData()) {
        push @result, FetchOneColumn();
    }

    return wantarray ? @result : $result[0];

}

#
#  last_insert_id()
#
#  Returns the integer ID created by the last INSERT into a table with an
#  auto_increment column.
#

sub last_insert_id {
    send_sql('SELECT LAST_INSERT_ID()');
    return FetchOneColumn();
}

#
#  read_custom_fields(@field_id)
#
#  Creates a new field hash for each of the fields enumerated in @field_id.
#  Returns a list of the new hashes.
#

sub read_custom_fields {
    my @field_id = @_;
    my @fields;

    my $field_set = join(',', @field_id);

    send_sql(qq{
        SELECT field_id, field_name, display_name, field_type,
               required, selection_id
        FROM custom_fields
        WHERE field_id IN ($field_set)
    });

    while (MoreSQLData()) {
        push @fields, read_cf_data();
    }

    read_default_field_data(\@fields);
    read_selection_metadata(\@fields);

    return wantarray ? @fields : $fields[0];

}

#
#  read_group($group_id)
#
#  Sets up the context for the main menu template in the case that a group
#  (identified by the numeric ID $group_id) is to be displayed.
#

sub read_group {
    my $group_id = shift;

    my $group_name = simple_query(qq{
        SELECT group_name
        FROM cf_groups
        WHERE group_id = $group_id
    });

    my @fields = read_group_data($group_id);

    read_selection_metadata(\@fields);

    $vars->{scope}           = 'group';
    $vars->{scope_id}        = $group_id;
    $vars->{scope_name}      = $group_name;
    $vars->{custom_fields}   = \@fields;
    $vars->{editable_fields} = \@fields;

}

#
#  read_group_data($group_id, %param)
#
#  Creates a field hash for every field in the group identified by the
#  numeric ID $group_id.  Returns a list containing all of the newly-created
#  field hashes.  %param is a parameter hash which is passed to
#  read_default_field_data().
#

sub read_group_data {
    my ($group_id, %param) = @_;
    my @fields;

    send_sql(qq{
        SELECT cf.field_id, cf.field_name, cf.display_name, cf.field_type,
               cf.required, cf.selection_id
        FROM custom_fields cf
          INNER JOIN cf_group_members gm USING (field_id)
        WHERE gm.group_id = $group_id
        ORDER BY gm.sortkey
    });

    while (MoreSQLData()) {
        push @fields, read_cf_data();
    }

    read_default_field_data(\@fields, \%param);

    return @fields;

}

#
#  read_product($product_id)
#
#  Sets up the context for the main menu template in the case that a product
#  (idenfitied by the numeric ID $product_id) is to be displayed.
#

sub read_product {
    my $product_id = shift;
    my ($last_name, $current_group, @fields, @groups);

    my $product_name = simple_query(qq{
        SELECT name
        FROM products
        WHERE id = $product_id
    });

    my $product_clause = defined $product_id
        ? "product_id = $product_id"
        : "product_id IS NULL";

    #  The left joins here are necessary because we still want to know
    #  when a product contains a group even if the group itself has no
    #  members:

    send_sql(qq{
        SELECT cf.field_id, cf.field_name, cf.display_name, cf.field_type,
               cf.required, cf.selection_id, g.group_id, g.group_name
        FROM cf_groups g
          INNER JOIN cf_product_members pm USING (group_id)
          LEFT JOIN cf_group_members gm USING (group_id)
          LEFT JOIN custom_fields cf USING (field_id)
        WHERE pm.$product_clause
        ORDER BY pm.sortkey, gm.sortkey
    });

    while (MoreSQLData()) {
        my ($field, $group_id, $group_name) = read_cf_data();

        push @fields, $field if defined $field;

        unless (defined $last_name
                && defined $group_name
                && $last_name eq $group_name) {
            $current_group = [ $field ? $field : () ];
            push @groups, {
                id => $group_id,
                name => $group_name,
                members => $current_group
            };
        } else {
            push @$current_group, $field;
        }

        $last_name = $group_name;

    }

    read_default_field_data(\@fields);
    read_selection_metadata(\@fields);

    my @editable = map @{ $$_{members} }, grep !defined $$_{name}, @groups;

    my @other_groups;
    my @group_ids = map $$_{id}, grep defined $$_{name}, @groups;
    my $group_id_clause;
    if (@group_ids) {
        my $group_set = join(',', @group_ids);
        $group_id_clause = "AND group_id NOT IN ($group_set)";
    } else {
        $group_id_clause = "";
    }

    process_query(qq{
        SELECT group_id, group_name
        FROM cf_groups
        WHERE group_name IS NOT NULL $group_id_clause
        ORDER BY group_name
    }, sub {
        push @other_groups, { id => $_[0], name => $_[1] };
    });

    $vars->{scope}               = 'product';
    $vars->{scope_id}            = $product_id;
    $vars->{scope_name}          = $product_name;
    $vars->{custom_fields}       = \@fields;
    $vars->{editable_fields}     = \@editable;
    $vars->{custom_field_groups} = \@groups;
    $vars->{other_groups}        = \@other_groups;

}

#
#  read_default_field_data($fields, $param)
#
#  This routine stores the default value for each of the fields in the array
#  referenced by $fields into each field's 'default' element.
#

sub read_default_field_data {
    my ($fields, $param) = @_;
    my %by_class = classify_fields(@$fields);

    while (my ($class, $fields) = each %by_class) {
        my @id = map $$_{id}, @$fields;
        my @values = read_field_data(undef, $class, \@id, $param);

        for (my $i = 0; $i < @values; ++$i) {
            $fields->[$i]{default} = $values[$i];
        }

        #  Special case for dates:
        if ($class eq 'date') {
            for my $field (grep !$$_{extended}, @$fields) {
                $field->{default} =~ s/ .*//;  # nuke time portion
                #  NOTE!  Here it is assumed that the time portion
                #  always follows the first space in a formatted date.
                #  Obviously this is may not always be the case.
                #  A more general mechanism is needed.
            }
        }

    }
}

#
#  read_field_data($bug_id, $class, $field_ids, $param)
#
#  This routine reads field data for the bug $bug_id for all fields of a
#  given class, namely $class.  $field_ids is a reference to an array of
#  numeric field IDs.
#

sub read_field_data {
    my ($bug_id, $class, $field_ids, $param) = @_;
    my $info = $classinfo{$class};
    my ($table, $column, $multiple) = @$info{qw(table column multiple)};
    my $field_set = join(',', @$field_ids);
    my %value;

    $column = $column->($param) if ref $column eq 'CODE';

    @value{ @$field_ids } = map [ ], 0 .. $#$field_ids;

    process_query(qq{
        SELECT field_id, $column
        FROM $table
        WHERE field_id IN ($field_set)
    }, sub {
        push @{ $value{ $_[0] } }, $_[1];
    });

    unless ($multiple) {
        while (my $key = each %value) {
            $value{$key} = $value{$key}[0];
        }
    }

    return @value{ @$field_ids };

}

#
#  read_selection_metadata($fields)
#
#  This routine stores information specific to selection fields in each of
#  the selection fields in the array referenced by $fields.  The extra
#  elements are 'unset' (a field's "unset" label), 'labels' (a mapping from
#  label_id to the text and order key of that label), and 'domain' (an array
#  of all labels in the field's domain, sorted into the proper order).  In
#  addition, the 'default' element of each field hash is reset to refer to
#  each labels' text, rather than their numeric IDs.
#

sub read_selection_metadata {
    my $fields = shift;
    my %by_class = classify_fields(@$fields);
    return unless defined($fields = $by_class{selection});

    my (%seen, %unset, %labels);

    my @sel_ids = grep !$seen{$_}++, map $$_{selection_id}, @$fields;
    my $id_set  = join(',', @sel_ids);

    process_query(qq{
        SELECT selection_id, unset_label
        FROM cf_selections
        WHERE selection_id IN ($id_set)
    }, sub {
        $unset{ $_[0] } = $_[1];
    });

    process_query(qq{
        SELECT selection_id, label_id, label, sortkey
        FROM cf_selection_labels
        WHERE selection_id IN ($id_set)
    }, sub {
        $labels{ $_[0] }{ $_[1] } = { text => $_[2], order => $_[3] };
    });

    for my $field (@$fields) {
        my ($selection_id, $default) = @$field{qw(selection_id default)};
        my $unset  = $field->{unset}  =  $unset{$selection_id};
        my $labels = $field->{labels} = $labels{$selection_id};
        $field->{domain} = [
            map  { $_->{text} }
            sort { $a->{order} <=> $b->{order} }
            values %$labels
        ];
        @$default = map  { $_->{text} }
                    sort { $a->{order} <=> $b->{order} }
                    @$labels{ @$default };
    }

}

#
#  reorder_members($scope, $id)
#
#  This routine returns an object which can be used to modify the members of
#  a product (if $scope is 'product') or group (if $scope is 'group').  $id
#  is the numeric ID of the product or group which is to be modified.
#
#  The returned object is a hash of subroutines, so one calls method "foo"
#  on it like this:  $object->{foo}->(@args).  The valid methods are:
#
#  index($id)
#  Returns the numeric index (starting at zero) of the group or field with
#  numeric ID $id.
#
#  remove($n)
#  Removes the group or field located at index $n.
#
#  insert($n, $id)
#  Inserts a new group or field with numeric ID $id before the existing
#  element at index $n.
#
#  commit()
#  Updates the database to reflect the new group or field order.  No further
#  methods should be called on the object once commit() has been called.
#

sub reorder_members {
    my ($scope, $id) = @_;
    my ($table, $column, $key, @member, %index, %sortkey);

    if ($scope eq 'product') {
        $table  = 'cf_product_members';
        $column = 'group_id';
        $key    = 'product_id';
    } elsif ($scope eq 'group') {
        $table  = 'cf_group_members';
        $column = 'field_id';
        $key    = 'group_id';
    } else {
        die qq(Internal error in reorder_members: unknown scope "$scope"\n);
    }

    process_query(qq{
        SELECT $column, sortkey
        FROM $table
        WHERE $key = $id
        ORDER BY sortkey
    }, sub {
        push @member, $_[0];
        $sortkey{ $_[0] } = $_[1];
    });

    @index{ @member } = 0 .. $#member;

    return {
        index  => sub { return $index{ $_[0] } },
        remove => sub { splice @member, $_[0], 1 },
        insert => sub { splice @member, $_[0], 0, $_[1] },
        commit => sub {
            my @when;
            for (my $i = 0; $i < @member; ++$i) {
                if ($sortkey{ $member[$i] } != $i + 1) {
                    push @when, "WHEN $member[$i] THEN " . ($i + 1);
                }
            }
            if (@when) {
                my $case_body = join(' ', @when) . ' ELSE sortkey';
                send_sql(qq{
                    UPDATE $table
                    SET sortkey = CASE $column $case_body END
                    WHERE $key = $id
                });
            }
        }
    };

}

#
#  read_field_metadata(@fields)
#
#  Returns an object that can be used to obtain information about each field
#  whose numeric ID appears in @fields.  The object is a
#  hash-of-subroutines, like the one returned by reorder_members().
#
#  each_table()
#  Iterates over each table which stores data for @fields.  Called in scalar
#  context, returns the name of one of these tables; in list context,
#  returns a second scalar, a reference to an array containing the field
#  IDs of each field for which data is stored in the table named by the
#  first element.
#
#  selection_id(@field_id)
#  Returns a list containing the selection IDs for each field in @field_ids.
#  @field_ids should be a subset of the fields initially passed to
#  read_field_metadata().
#

sub read_field_metadata {
    my @fields = @_;
    my (%by_table, %selection_id);

    if (@fields) {
        my $field_set = join(',', @fields);
        process_query(qq{
            SELECT field_id, field_type, selection_id
            FROM custom_fields
            WHERE field_id IN ($field_set)
        }, sub {
            my ($field_id, $type, $sel_id) = @_;
            defined(my $class = $class{$type})
                or die qq(Internal error: unknown type "$type" in database\n);
            my $table = $classinfo{$class}{table};
            push @{ $by_table{$table} }, $field_id;
            $selection_id{$field_id} = $sel_id;
        });
    }

    return {
        each_table   => sub { each %by_table },
        selection_id => sub { @selection_id{ @_ } }
    };

}

#
#  read_cf_data()
#
#  Creates a field hash and returns it.  A SELECT statement must previously
#  have been sent to the database which returns field information; the
#  first through sixth columns must be an ID, a name, a display name, a
#  type, a "required" flag, and a selection ID, in that order.  Any extra
#  columns are returned after the new field hash in the return list.  If the
#  first column is NULL, undef is returned.
#

sub read_cf_data {
    my ($id, $name, $dname, $type, $required, $selection_id, @extra) =
        FetchSQLData();
    my $field;

    if (defined $id) {
        defined(my $class = $class{$type})
            or die qq(Database corrupted: unknown type "$type"\n);
        $field = {
            id           => $id,
            name         => $name,
            display_name => $dname,
            type         => $type,
            required     => $required,
            extended     => $type eq 'l' || $type eq 'm' || $type eq 't',
            selection_id => $selection_id,
            is_integer   => $class eq 'integer',
            is_date      => $class eq 'date',
            is_string    => $class eq 'l_string' || $class eq 's_string',
            is_selection => $class eq 'selection',
        };
    }

    return wantarray ? ($field, @extra) : $field;

}

#
#  parse_integer($str)
#
#  Attempts to parse a string containing an integer.  Leading and trailing
#  whitespace is trimmed; what is left must resemble the usual integer
#  pattern, otherwise an exception is thrown.  The trimmed string is
#  returned.
#

sub parse_integer {
    my $str = shift;

    $str = trim($str);

    if ($str eq "") {
        $str = 'NULL';
    } elsif ($str !~ /[-+]?\d+$/) {
        die qq(Invalid integer value "$str"\n);
    }

    return $str;

}

#
#  parse_date($str, $withtime)
#
#  Attempts to parse a string containing a date.  Leading and trailing
#  whitespace is trimmed; what is left must be parseable by Date::Parse, or
#  an exception is thrown.  Otherwise, the parsed date is cast in the format
#  YYYYMMDDHHMMSS, which is returned.  If the original string contained no
#  non-whitespace characters, the string 'NULL' is returned instead.
#

sub parse_date {
    my ($str, $withtime) = @_;

    $str = trim($str);

    if ($str eq "") {
        $str = 'NULL';
    } else {
        my ($sec, $min, $hour, $day, $month, $year) = strptime($str)
            or die qq(Invalid date value "$str"\n);
        $sec = $min = $hour = 0 if !$withtime;
        $str = sprintf '%04d%02d%02d%02d%02d%02d',
                       $year + 1900, $month + 1, $day, $hour, $min, $sec;
    }

    return $str;

}

#
#  classify_fields(@fields)
#
#  Iteratates over the fields in @fields, storing each one in a hash keyed
#  by the class of each.  Returns the flattened hash.
#

sub classify_fields {
    my @fields = @_;
    my %by_class;

    for my $field (@fields) {
        my $class = $class{ $field->{type} };
        push @{ $by_class{$class} }, $field;
    }

    return %by_class;
}

#
#  date_column($param)
#
#  This routine returns an expression which is used to format a date column
#  fetched from the database.  $param is a reference to a named parameter
#  hash.  If the hash has an element 'date_format', that element is used as
#  the format string for a call to DATE_FORMAT().  If the element is null,
#  the unformatted "value" column is returned instead.  If $param does not
#  have a 'date_format' element, the date is formatted according to the
#  global $DEFAULT_DATE_FORMAT format.
#

sub date_column {
    my $param = shift;

    my $format = exists $param->{date_format}
        ? $param->{date_format}
        : $DEFAULT_DATE_FORMAT;

    my $expr = defined $format ? "DATE_FORMAT(value, '$format')" : "value";

    return $expr;

}

